// Much of this was based off of Ben Rudiak-Gould's huffyuv source code

#ifndef _MAIN_HEADER
#define _MAIN_HEADER

#include <process.h>
#include <malloc.h>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>
#include <malloc.h>
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <stdio.h>

//#define X64_BUILD // this will be set automaticly if using the release 64 configuration

//#define LOSSLESS_CHECK			// enables decompressing encoded frames and comparing to original

#define LAGARITH_RELEASE		// if this is a version to release, disables all debugging info 

#ifndef LAGARITH_RELEASE
#define TRY_CATCH_FUNC			// enables try/catch macro

inline void crash(){ // force an error to provide an easy way to attach the debugger
#ifndef X64_BUILD
	__asm{
		int 3
	}
#endif
}

inline void infinite_loop(){ // force an infinite loop to provide an easy way to attach the debugger when crash doesn't work
#ifndef X64_BUILD
	__asm{ // don't have to worry about assembly being optimized out
ZERO_ECX:
		xor		ecx,ecx
		jcxz	ZERO_ECX
	}
#else
		while(true){};
#endif
}

#endif

#ifndef LAGARITH_RELEASE
#define bang(str) MessageBox(HWND_DESKTOP, str, "Error", MB_OK | MB_ICONEXCLAMATION)
#else
#define bang(str)
#endif

#ifdef TRY_CATCH_FUNC
#define try_catch( f ) \
try { f }\
catch ( char * tc_cmsg){\
char * tc_msg = (char *)malloc(strlen(tc_cmsg)+128);\
sprintf_s(tc_msg,strlen(tc_cmsg)+128"Exception passed up to %s, line %d.\nOriginal exception: %s\n", __FILE__, __LINE__,tc_cmsg);\
MessageBox (HWND_DESKTOP, tc_msg, "Error", MB_OK | MB_ICONEXCLAMATION);\
throw tc_msg;\
}\
catch (...){\
char * tc_msg = (char*)malloc(128);\
sprintf_s(tc_msg,128,"Exception caught in %s, line %d", __FILE__, __LINE__);\
MessageBox (HWND_DESKTOP, tc_msg, "Error", MB_OK | MB_ICONEXCLAMATION); \
throw tc_msg;\
}
#else
#define try_catch( f ) f
#endif

inline void * lag_aligned_malloc( void *ptr, int size, int align, char *str ) {
	if ( ptr ){
		try {
			#ifndef LAGARITH_RELEASE
			char msg[128];
			sprintf_s(msg,128,"Buffer '%s' is not null, attempting to free it...",str);
			MessageBox (HWND_DESKTOP, msg, "Error", MB_OK | MB_ICONEXCLAMATION);
			#endif
			_aligned_free(ptr);
		} catch ( ... ){
			#ifndef LAGARITH_RELEASE
			char msg[256];
			sprintf_s(msg,128,"An exception occurred when attempting to free non-null buffer '%s' in lag_aligned_malloc",str);
			MessageBox (HWND_DESKTOP, msg, "Error", MB_OK | MB_ICONEXCLAMATION);
			#endif
		}
	}
	return _aligned_malloc(size,align);
}

#ifdef LAGARITH_RELEASE
#define lag_aligned_free(ptr, str) { \
	if ( ptr ){ \
		try {\
			_aligned_free(ptr);\
		} catch ( ... ){ } \
	} \
	ptr=NULL;\
}
#else 
#define lag_aligned_free(ptr, str) { \
	if ( ptr ){ \
		try { _aligned_free(ptr); } catch ( ... ){\
			char err_msg[256];\
			sprintf_s(err_msg,256,"Error when attempting to free pointer %s, value = 0x%X - file %s line %d",str,ptr,__FILE__, __LINE__);\
			MessageBox (HWND_DESKTOP, err_msg, "Error", MB_OK | MB_ICONEXCLAMATION);\
		} \
	} \
	ptr=NULL;\
}
#endif

// y must be 2^n
#define align_round(x,y) ((((unsigned int)(x))+(y-1))&(~(y-1)))

#include "resource.h"
#include "compact.h"

//#pragma hdrstop

static const DWORD FOURCC_LAGS = mmioFOURCC('L','A','G','S');
static const DWORD FOURCC_YUY2 = mmioFOURCC('Y','U','Y','2');
static const DWORD FOURCC_UYVY = mmioFOURCC('U','Y','V','Y');
static const DWORD FOURCC_YV12 = mmioFOURCC('Y','V','1','2');
static const DWORD FOURCC_HDYC = mmioFOURCC('H','D','Y','C'); // Blackmagic Intensity
															  //  (HDMI capture card)
// possible frame flags

#define UNCOMPRESSED		1	// Used for debugging
#define UNALIGNED_RGB24		2	// RGB24 keyframe with a width that is not a multiple of 4; this has to be handled 
								// differently due to the fact that previously, the codec did not remove byte padding to 
								// align each scan line. Old versions are decoded as ARITH_RGB24
#define ARITH_YUY2			3	// Standard YUY2 keyframe frame
#define ARITH_RGB24			4	// Standard RGB24 keyframe frame
#define BYTEFRAME			5	// solid greyscale color frame
#define PIXELFRAME			6	// solid non-greyscale color frame
#define OBSOLETE_ARITH		7	// Old RGB keyframe frame, kept only to for backwards compatablity
#define ARITH_ALPHA			8	// Standard RGBA keyframe frame
#define PIXELFRAME_ALPHA	9	// RGBA pixel frame
#define ARITH_YV12			10	// Standard YV12 frame	
#define REDUCED_RES         11	// Reduced Resolution frame, for bait & swap editing.

//#define INTERLACED_FLAG		64	// Set in the frame type byte if video is interlaced

// possible colorspaces
#define RGB24	24
#define RGB32	32
#define YUY2	16
#define YV12	12

struct threadinfo {
	volatile const unsigned char * source;	// data source
	volatile unsigned char * dest;		// data destination
	unsigned char * buffer;	// buffer used for median prediction or restoration
	unsigned int SSE2;
	unsigned int SSE;
	unsigned int width;
	unsigned int height;
	unsigned int format;
	volatile unsigned int length;	// uncompressed data length
	volatile unsigned int size;		// compressed data length
	unsigned int lum;				// needed for YUY2 prediction
	HANDLE mutex;
	char mutex_name[12];
#ifdef _DEBUG
	char * name;
#endif
	HANDLE thread;
	CompressClass cObj;
};

class CodecInst {
public:
	unsigned char * buffer;
	unsigned char * prev;
	const unsigned char * in;
	unsigned char * out;
	unsigned char * buffer2;
//	unsigned char * lossy_buffer;
	unsigned int length;
	unsigned int width;
	unsigned int height;
	unsigned int format;	//input format for compressing, output format for decompression. Also the bitdepth.
	bool nullframes;
	bool use_alpha;
//	bool interlaced;
	int lossy_option;
	bool multithreading;
	int started;			//if the codec has been properly initalized yet
	threadinfo info_a;
	threadinfo info_b;
	threadinfo info_c;
	CompressClass cObj;

#ifndef X64_BUILD
	int SSE2;	
	int SSE;
	int MMX;
#endif

	CodecInst();
	~CodecInst();

	DWORD GetState(LPVOID pv, DWORD dwSize);
	DWORD SetState(LPVOID pv, DWORD dwSize);
	DWORD Configure(HWND hwnd);
	DWORD GetInfo(ICINFO* icinfo, DWORD dwSize);

	DWORD CompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
	DWORD CompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
	DWORD CompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
	DWORD CompressGetSize(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
	DWORD Compress(ICCOMPRESS* icinfo, DWORD dwSize);
	DWORD CompressEnd();

	DWORD DecompressQuery(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
	DWORD DecompressGetFormat(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
	DWORD DecompressBegin(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
	DWORD Decompress(ICDECOMPRESS* icinfo, DWORD dwSize);
	DWORD DecompressGetPalette(LPBITMAPINFOHEADER lpbiIn, LPBITMAPINFOHEADER lpbiOut);
	DWORD DecompressEnd();

	BOOL QueryConfigure();

	void uncompact_macro( const unsigned char * _in, unsigned char * _out, unsigned int _length, unsigned int _width, unsigned int _height, threadinfo * _thread, int _format);
	int InitThreads( int encode);
	void EndThreads();

	int CompressRGB24(ICCOMPRESS* icinfo);
	int CompressRGBA(ICCOMPRESS* icinfo);
	int CompressYUY2(ICCOMPRESS* icinfo);
	int CompressYV12(ICCOMPRESS* icinfo);
	int CompressLossy(ICCOMPRESS* icinfo);
	int CompressReduced(ICCOMPRESS* icinfo);
	bool SolidFrameTestRGB24(ICCOMPRESS* icinfo);
	bool SolidFrameTestRGB32(ICCOMPRESS* icinfo);

	void ByteDecompress();
	void PixelDecompress();
	void AlphaPixelDecompress();
	void ArithDecompress();
	void ArithRGBADecompress();
	void ArithYUY2Decompress();
	void ArithYV12Decompress();
	void UnalignedDecompress();
	void ReduceResDecompress();
};

CodecInst* Open(ICOPEN* icinfo);
DWORD Close(CodecInst* pinst);

#endif