#include "lagarith.h"
#include "prediction.h"

#define lag_ResumeThread(x) {\
	while( ResumeThread(x) != 1){\
		Sleep(0);\
	}\
}


DWORD WINAPI encode_worker_thread( LPVOID i ){

	threadinfo * info = (threadinfo *)i;
	const unsigned int width = info->width;
	const unsigned int height = info->height;
#ifndef X64_BUILD
	const unsigned int SSE2 = info->SSE2;
	const unsigned int SSE = info->SSE;
	const unsigned int stride = align_round(width,(SSE2?16:8));
#else
	const unsigned int stride = align_round(width,16);
#endif
	const unsigned char * src=NULL;
	unsigned char * dest=NULL;
	unsigned char * const buffer = (unsigned char *)info->buffer;
	const unsigned int format=info->format;
	const unsigned int lum = info->lum;
	while ( info->length != 0xFFFFFFFF ){

		src = (const unsigned char *)info->source;
		dest = (unsigned char *)info->dest;

		unsigned char * dst;
		if (  width != stride ){
			dst=(unsigned char *)align_round(dest,16);
		} else {
			dst=buffer;
		}

#ifndef X64_BUILD
		if ( format != YUY2 ){
			if ( SSE2 ){
				SSE2_BlockPredict(src,dst,stride,stride*height,(format!=YV12));
			} else {
				MMX_BlockPredict(src,dst,stride,stride*height,SSE,(format!=YV12));
			}
		} else {
			if ( SSE2 ){
				SSE2_Predict_YUY2(src,dst,stride,height,lum);
			} else {
				MMX_Predict_YUY2(src,dst,stride,height,SSE,lum);
			}
		}
#else
		if ( format != YUY2 ){
			SSE2_BlockPredict(src,dst,stride,stride*height,(format!=YV12));
		} else {
			SSE2_Predict_YUY2(src,dst,stride,height,lum);
		}
#endif
		if ( width!= stride ){
			unsigned char * padded = dst;
			unsigned char * stripped = buffer;
			for ( unsigned int y=0;y<height;y++){
				memcpy(stripped+y*width,padded+y*stride,width);
			}
		}

		info->size=info->cObj.compact(buffer,dest,width*height);
		assert( *(__int64*)dest != 0 );
		info->length=0;
		SuspendThread(info->thread);// go to sleep, main thread will resume it
	}
	info->cObj.FreeCompressBuffers();
	lag_aligned_free(info->buffer,"Thread Buffer");
	info->length=0;
	return 0;
}

DWORD WINAPI decode_worker_thread( LPVOID i ){
	threadinfo * info = (threadinfo *)i;
	unsigned int width;
	unsigned int height;
	unsigned char * src=NULL;
	unsigned char * dest=NULL;
	unsigned int format;
	unsigned int length;
	while ( info->length != 0xFFFFFFFF ){
		src = (unsigned char *)info->source;
		dest = (unsigned char *)info->dest;

		length=info->length;
		format=info->format;
		width = info->width;
		height = info->height;

		info->cObj.uncompact(src,dest,length);
		if ( format != YUY2 ){
			ASM_BlockRestore(dest,width,width*height,format!=YV12);
		}

		info->length=0;
		SuspendThread(info->thread);
	}
	info->cObj.FreeCompressBuffers();
	lag_aligned_free(info->buffer,"Thread Buffer");
	info->length=0;
	return 0;
}

#define wait_for_threads(threads) { \
	if ( multithreading ) { \
		if ( threads == 2) { \
			while ( info_a.length || info_b.length ) \
				Sleep(0); \
		} else if ( threads == 3 ){ \
			while ( info_a.length || info_b.length || info_c.length) \
				Sleep(0); \
		} \
	} \
}

int CodecInst::InitThreads( int encode ){

	DWORD temp;
	info_a.length=0;
	info_b.length=0;
	info_c.length=0;

	unsigned int use_format=format;
	if ( lossy_option ){
		if ( lossy_option==1 ){
			use_format=YUY2;
		} else {
			use_format=YV12;
		}
	}

	info_a.width=width;
	if ( use_format > YUY2 ){
		info_b.width=width;
	} else {
		info_b.width=width/2;
	}
	info_c.width=width;

	info_a.height=height;
	if ( use_format != YV12 ){
		info_b.height=height;
	} else {
		info_b.height=height/2;
	}
	info_c.height=height;

	if ( lossy_option==3 ){
		info_a.width=width/2;
		info_a.height=height/2;
		info_b.width=width/4;
		info_b.height=height/4;
	}

	info_a.format=use_format;
	info_b.format=use_format;
	info_c.format=use_format;

#ifndef X64_BUILD
	info_a.SSE=SSE;
	info_b.SSE=SSE;
	info_c.SSE=SSE;

	info_a.SSE2=SSE2;
	info_b.SSE2=SSE2;
	info_c.SSE2=SSE2;
#endif

	info_a.thread=NULL;
	info_b.thread=NULL;
	info_c.thread=NULL;
	info_a.buffer=NULL;
	info_b.buffer=NULL;
	info_c.buffer=NULL;
#ifdef _DEBUG 
	info_a.name="Thread A";
	info_b.name="Thread B";
	info_c.name="Thread C";
#endif

	info_a.lum=1;
	info_b.lum=0;
	info_c.lum=0;

	int buffer_a = align_round(width,16)*(height)+2048;
	int buffer_b = align_round(info_b.width,16)*info_b.height+2048;

	if ( !info_a.cObj.InitCompressBuffers( buffer_a ) || !info_b.cObj.InitCompressBuffers( buffer_b ) 
		|| !(info_a.buffer=(unsigned char *)lag_aligned_malloc(info_a.buffer,buffer_a,16,"Info_a.buffer"))
		|| !(info_b.buffer=(unsigned char *)lag_aligned_malloc(info_b.buffer,buffer_b,16,"Info_b.buffer")) ){
		info_a.cObj.FreeCompressBuffers();
		info_b.cObj.FreeCompressBuffers();
		lag_aligned_free(info_a.buffer,"Info_a.buffer");
		lag_aligned_free(info_b.buffer,"Info_b.buffer");
		info_a.thread=NULL;
		info_b.thread=NULL;
		return ICERR_MEMORY;
	} else { 
		if ( format == RGB32 ){
			if ( !info_c.cObj.InitCompressBuffers( buffer_a ) 
				|| !(info_c.buffer=(unsigned char *)lag_aligned_malloc(info_c.buffer,buffer_a,16,"Info_c.buffer"))){
				info_a.cObj.FreeCompressBuffers();
				info_b.cObj.FreeCompressBuffers();
				info_c.cObj.FreeCompressBuffers();
				lag_aligned_free(info_a.buffer,"Info_a.buffer");
				lag_aligned_free(info_b.buffer,"Info_b.buffer");
				lag_aligned_free(info_c.buffer,"Info_c.buffer");
				info_a.thread=NULL;
				info_b.thread=NULL;
				info_c.thread=NULL;
				return ICERR_MEMORY;
			}
			if ( encode ){
				info_c.thread=CreateThread(NULL,0, encode_worker_thread, &info_c,CREATE_SUSPENDED,&temp);
			} else {
				info_c.thread=CreateThread(NULL,0, decode_worker_thread, &info_c,CREATE_SUSPENDED,&temp);
			}
		}
		if ( encode ){
			info_a.thread=CreateThread(NULL,0, encode_worker_thread, &info_a,CREATE_SUSPENDED,&temp);
			info_b.thread=CreateThread(NULL,0, encode_worker_thread, &info_b,CREATE_SUSPENDED,&temp);
		} else {
			info_a.thread=CreateThread(NULL,0, decode_worker_thread, &info_a,CREATE_SUSPENDED,&temp);
			info_b.thread=CreateThread(NULL,0, decode_worker_thread, &info_b,CREATE_SUSPENDED,&temp);
		}
		if ( !info_a.thread || !info_b.thread || (format==RGB32 && !info_c.thread )){
			info_a.cObj.FreeCompressBuffers();
			info_b.cObj.FreeCompressBuffers();
			if ( format==RGB32 ){
				info_c.cObj.FreeCompressBuffers();
				lag_aligned_free(info_c.buffer,"Info_c.buffer");
			}
			lag_aligned_free(info_a.buffer,"Info_a.buffer");
			lag_aligned_free(info_b.buffer,"Info_b.buffer");
			info_a.thread=NULL;
			info_b.thread=NULL;
			info_c.thread=NULL;
			return ICERR_INTERNAL;
		}
	}
	if ( use_format >= RGB24 ){
		SetThreadPriority(info_b.thread,THREAD_PRIORITY_ABOVE_NORMAL);
	} else {
		SetThreadPriority(info_a.thread,THREAD_PRIORITY_ABOVE_NORMAL);
	}
	return ICERR_OK;
}

void CodecInst::EndThreads(){
	info_a.length=0xFFFFFFFF;
	info_b.length=0xFFFFFFFF;
	info_c.length=0xFFFFFFFF;
	if ( info_a.thread ){
		lag_ResumeThread(info_a.thread);
	}
	if ( info_b.thread ){
		lag_ResumeThread(info_b.thread);
	}
	if ( info_c.thread && format == RGB32 ){
		lag_ResumeThread(info_c.thread);
	}
	while ( (info_a.length && info_a.thread) || (info_b.length && info_b.thread) || (info_c.length && format == RGB32 && info_c.thread) ){
		Sleep(1);
	}

	if ( !CloseHandle(info_a.thread) ){
		/*LPVOID lpMsgBuf;
		FormatMessage( 
			FORMAT_MESSAGE_ALLOCATE_BUFFER | 
			FORMAT_MESSAGE_FROM_SYSTEM | 
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			GetLastError(),
			0, // Default language
			(LPTSTR) &lpMsgBuf,
			0,
			NULL 
		);
		MessageBox( NULL, (char *)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION );
		LocalFree( lpMsgBuf );*/
		//MessageBox (HWND_DESKTOP, "CloseHandle failed for thread 0x0A", "Error", MB_OK | MB_ICONEXCLAMATION);
	}
	if ( !CloseHandle(info_b.thread) ){
	//	MessageBox (HWND_DESKTOP, "CloseHandle failed for thread 0x0B", "Error", MB_OK | MB_ICONEXCLAMATION);
	}
	if ( info_c.thread && format == RGB32 ){
		if ( !CloseHandle(info_c.thread) ){
			//MessageBox (HWND_DESKTOP,"CloseHandle failed for thread 0x0C", "Error", MB_OK | MB_ICONEXCLAMATION);
		}
	}
	info_a.thread=NULL;
	info_b.thread=NULL;
	info_c.thread=NULL;
	info_a.length=0;
	info_b.length=0;
	info_c.length=0;
}

//MessageBox (HWND_DESKTOP, msg, "Error", MB_OK | MB_ICONEXCLAMATION);